(in-package :domains)

(define (domain aibo-ball)
	(:requirements :conditional-effects :sensing :uncertainty)  

	(:action check-room
		:effect (and (observes (in-green-rm))
                         (not (facing-ball))
                         (not (facing-dest))
                         (know-room))
	)

	(:action check-holding
		:effect (and (observes (holding))
                         (not (facing-ball))
                         (not (facing-dest))
                         (know-holding))
	)
  
      (:action locate-ball
            ; Precondition is specified as conditional to force SGP
            ; to choose between locate-ball and release ball.
            :effect (and (not (facing-dest))
                         (when (not (holding)) (and (observes (ball-in-g-rm)) (facing-ball)))
                         (when (holding) (not (facing-ball)) (not (holding)))
                         (know-ball-room)
                         )
	)
  
	(:action grab-ball
		:precondition (and (facing-ball) (not (holding)))
            :effect (and (holding) (not (facing-ball)) (not (facing-dest))
                         (when (ball-in-g-rm) (in-green-rm))
                         (when (not (ball-in-g-rm)) (not (in-green-rm)))
                         )
	)
  
	(:action release-ball
            ; Precondition is specified as conditional so that SGP
            ; realizes that release-ball will undo locate-ball
            ; if (not (holding))
            :effect (and (not (facing-dest))
                         (when (holding) (and (not (holding)) (facing-ball)))
                         (when (not (holding)) (not (facing-ball)))
                         )
	)

	(:action face-dest
		:effect (and (facing-dest) (not (facing-ball)))
	)

	(:action go-dest
		:precondition (facing-dest)
		:effect (and (when (and (in-green-rm) (holding))
				(and (not (in-green-rm)) (not (ball-in-g-rm))
                             (not (facing-dest))))
			     (when (and (in-green-rm) (not (holding)))
				(and (not (in-green-rm)) (not (facing-dest))
                             (not (facing-ball))))
			     (when (and (not (in-green-rm)) (holding))
				(and (in-green-rm) (ball-in-g-rm)
                             (not (facing-dest))))
			     (when (and (not (in-green-rm)) (not (holding)))
                        (and (in-green-rm) (not (facing-dest))
                             (not (facing-ball))))
			)
          )
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; End aibo-ball domain definition. Below are problem definitions ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Cheat to get around name spaces conveniently
(defvar y 1)
(defvar n -1)
(defvar u 0)


(defmacro defprob (name rrm brm h)
  `(define (problem ,name) (:domain aibo-ball)
       (:init (not (facing-ball))
              (not (facing-dest))
              (oneof 
              ,@(genstate rrm brm h)
               )
              )
       (:goal (and (in-green-rm) (ball-in-g-rm)))
     )
  )

(defun genstate (rrm brm h)
  (let* ((r-pos (cond 
                 ((= rrm U) '(Y N))
                 ((= rrm Y) '(Y))
                 ((= rrm N) '(N))
                 ))
         (b-pos (cond 
                 ((= brm U) '(Y N))
                 ((= brm Y) '(Y))
                 ((= brm N) '(N))
                 ))
         (h-pos (cond 
                 ((= h U) '(Y N))
                 ((= h Y) '(Y))
                 ((= h N) '(N))
                 ))
         ) ; End let* definitions
    (permute-r h-pos b-pos r-pos ())
    )
  )

(defun permute-r (h-pos b-pos r-pos retval)
  (if (null r-pos)
      retval ; Done
      (let* ((r-val (car r-pos))
             (b-rpos (if (equal h-pos '(Y)) ; If we're holding the ball
                         (list r-val) ; It must be in the same room as us
                         b-pos
                       ))
             (newworlds (permute-b h-pos b-rpos r-val ())))
        (permute-r h-pos b-pos (cdr r-pos)
                   (append newworlds retval))
        )
    )
  )

(defun permute-b (h-pos b-pos r-val retval)
  (if (null b-pos)
      retval ; Done
      (let* ((b-val (car b-pos))
             (h-rpos (if (or (and (char= r-val 'y) (char= b-val 'n))
                             (and (char= r-val 'n) (char= b-val 'y)))
                         '(n) ; y would be unphysical so don't allow it
                         h-pos ; h can take on either value physically
                       ))
             (newworlds (permute-h h-rpos b-val r-val ())))
        (permute-b h-pos (cdr b-pos) r-val
                   (append newworlds retval))
        )
    )
  )

(defun permute-h (h-pos b-val r-val retval)
  (if (null h-pos)
      retval ; Done
      (let ((world (makestateoneof r-val b-val (car h-pos))))
        (permute-h (cdr h-pos) b-val r-val (cons world retval))
        )
    )
  )

(defun makestateoneof (r b h)
  `(and ,(if (char= r 'Y)
                     '(in-green-rm)
                     '(not (in-green-rm))
                   )
                ,(if (char= b 'Y)
                     '(ball-in-g-rm)
                     '(not (ball-in-g-rm))
                   )
                ,(if (char= h 'Y)
                     '(holding)
                     '(not (holding))
                   )
        )
  )




;;; Problems are named by three bits with yes (Y), no (N) or unknown (U)
;;; 1) in-green-rm 2) ball-in-g-rm 3) holding

(define (problem yyy) (:domain aibo-ball)
	(:init (in-green-rm)
		 (ball-in-g-rm)
		 (holding)
	)
	(:goal (and (in-green-rm) (ball-in-g-rm)))
)


;; No unphysical nyy or yny domains since we can't hold the ball
;; when its in the other room.

(define (problem yyn) (:domain aibo-ball)
	(:init (in-green-rm)
		 (ball-in-g-rm)
		 (not (holding))
	)
	(:goal (and (in-green-rm) (ball-in-g-rm)))
)

(define (problem nny) (:domain aibo-ball)
	(:init (not (in-green-rm))
		 (not (ball-in-g-rm))
		 (holding)
	)
	(:goal (and (in-green-rm) (ball-in-g-rm)))
)

(define (problem nyn) (:domain aibo-ball)
	(:init (not (in-green-rm))
		 (ball-in-g-rm)
		 (not (holding))
	)
	(:goal (and (in-green-rm) (ball-in-g-rm)))
)

(define (problem ynn) (:domain aibo-ball)
	(:init (in-green-rm)
		 (not (ball-in-g-rm))
		 (not (holding))
	)
	(:goal (and (in-green-rm) (ball-in-g-rm)))
)

(define (problem nnn) (:domain aibo-ball)
	(:init (not (in-green-rm))
		 (not (ball-in-g-rm))
		 (not (holding))
	)
	(:goal (and (in-green-rm) (ball-in-g-rm)))
  )

(define (problem YUU) (:domain aibo-ball)
	(:init (in-green-rm)
             (oneof (and (ball-in-g-rm) (holding))
                    (and (ball-in-g-rm) (not (holding)))
                    (and (not (ball-in-g-rm)) (not (holding)))
                    )
        )
	(:goal (and (in-green-rm) (ball-in-g-rm)))
  )

(define (problem NUU) (:domain aibo-ball)
	(:init (not (in-green-rm))
             (oneof (and (not (ball-in-g-rm)) (holding))
                    (and (not (ball-in-g-rm)) (not (holding)))
                    (and (ball-in-g-rm) (not (holding)))
                    )
        )
	(:goal (and (in-green-rm) (ball-in-g-rm)))
  )

(define (problem UYU) (:domain aibo-ball)
	(:init (ball-in-g-rm)
             (oneof (and (in-green-rm) (holding))
                    (and (in-green-rm) (not (holding)))
                    (and (not (in-green-rm)) (not (holding)))
                    )
        )
	(:goal (and (in-green-rm) (ball-in-g-rm)))
  )

(define (problem UNU) (:domain aibo-ball)
	(:init (not (ball-in-g-rm))
             (oneof (and (not (in-green-rm)) (holding))
                    (and (not (in-green-rm)) (not (holding)))
                    (and (in-green-rm) (not (holding)))
                    )
        )
	(:goal (and (in-green-rm) (ball-in-g-rm)))
  )

(define (problem UUY) (:domain aibo-ball)
	(:init (holding)
             (oneof (and (in-green-rm) (ball-in-g-rm))
                    (and (not (in-green-rm)) (not (ball-in-g-rm)))
                    )
        )
	(:goal (and (in-green-rm) (ball-in-g-rm)))
  )

(define (problem UUN) (:domain aibo-ball)
	(:init (not (holding))
             (oneof (and (in-green-rm) (ball-in-g-rm))
                    (and (not (in-green-rm)) (ball-in-g-rm))
                    (and (in-green-rm) (not (ball-in-g-rm)))
                    (and (not (in-green-rm)) (not (ball-in-g-rm)))
                    )
        )
	(:goal (and (in-green-rm) (ball-in-g-rm)))
  )

;; Real world completely unknown problem
(define (problem UUU) (:domain aibo-ball)
	(:init (not (facing-ball))
	       (not (facing-dest))
	       (oneof (and (in-green-rm) (ball-in-g-rm) (holding))
		      (and (in-green-rm) (ball-in-g-rm) (not (holding)))
		      (and (in-green-rm) (not (ball-in-g-rm)) (not (holding)))
		      (and (not (in-green-rm)) (ball-in-g-rm) (not (holding)))
		      (and (not (in-green-rm)) (not (ball-in-g-rm)) (holding))
		      (and (not (in-green-rm)) (not (ball-in-g-rm)) (not (holding)))
		)
	)
	(:goal (and (in-green-rm) (ball-in-g-rm)))
)