(in-package :cg-user)

;; Cheat to get around name spaces conveniently
(defvar y 1)
(defvar n -1)
(defvar u 0)

(defvar cur-rrm U)
(defvar cur-brm U)
(defvar cur-h   U)

(defmacro getplan (rrm brm h)
  `(progn
     (domains::defprob domains::isrunprob ,(eval rrm) ,(eval brm) ,(eval h))
     (sgp::plan 'isrunprob)
     )
  )
             
(defun run ()
  (format t "~%~%=== Starting the planner with state UUN ===~%")
  (let ((plan (getplan U U N)) (world (getworlds 0 0 -1)))
    (run-aux plan world)
    )
  )

(defun run-aux (plan worlds)
  (if (or (null plan) (null (car plan)))
      (progn
        (command done)
        nil)
    ; Else
    ;; In the process of binding newworlds we exicute
    ;; the actions on the current level.
    (let ((newworlds (processlevel (car plan) worlds)))
      (if (equal newworlds r)
          (progn
            (format t "~%Replan requested so generating a new plan from ~A~A~A~%"
              cur-rrm cur-brm cur-h)
            (command "rp")
            (let ((plan (getplan cur-rrm cur-brm cur-h))
                  (worlds (getworlds cur-rrm cur-brm cur-h)))
              (run-aux plan worlds))
            )
          (run-aux (cdr plan) newworlds)
        )
      )
    )
  )

(defvar last-action nil)

(defun processlevel (lvl worlds)
  (cond
   
   ((null lvl) worlds) ; Nothing to do here
                       ; So just return the existing worlds
   
   ; Special case where we need a replan, pass the need for
   ; re planning back up the chain.
   ((equal worlds r) r)
   
   (t (progn   
        ; Display a progress report
        (format t "The current plan level is: ~A~%" lvl)
        (format t "And the possible worlds are: ~A~%" worlds)
        
        (let ((next-step (car lvl)))
          (if (and (not (null next-step))
                   (not (equal (getaction next-step) last-action))
                   (should-do next-step worlds))
              ;; If we should exicute the next step given the worlds
              ;; we might be in, do it and recurse along the level.
              (progn
                (setf last-action (getaction next-step))
                (processlevel (cdr lvl) (exicute (getaction next-step) worlds))
                )
              ;; Otherwise, don't exicute but do recurse along level.
              (processlevel (cdr lvl) worlds)
            )
          
          )
        ) 
      )
   )
  )

(defun get-replan-data ()
  ;; If a replan has been requested the next
  ;; three characters will be information about
  ;; the new world needed to generate the new
  ;; plan.
  (list
   (listenSock)
   (listenSock)
   (listenSock)
   )
  )

(defun should-do (step worlds)
  (let ((noex-wrlds (get-noex-wrlds step))
        (pos-wrlds (mapcar 'car worlds)))
    (check-noex-wrlds noex-wrlds pos-wrlds)
    )
  )

(defun check-noex-wrlds (noex pos)
  (cond
   ((null noex) t)
   ((member (car noex) pos) nil)
   (t (check-noex-wrlds (cdr noex) pos))
   )
  )

(defun exicute (action current-wrlds)
  (format t "Exicuting the ~A action.~%" action)
  
  ;; This entier function consists mainly of one big
  ;; switch over what action is being exicuted.
  (cond
   ((equal action 'domains::check-room) ; Sences in-green-rm
    (prune-worlds (setf cur-rrm (command inGreenRm)) 0 0 current-wrlds))
   
   ((equal action 'domains::check-holding) ; Senses holding
    (prune-worlds 0 0 (setf cur-h (command holdingBall)) current-wrlds))
   
   ((equal action 'domains::locate-ball) ; Senses ball-in-g-rm
    (prune-worlds 0 (setf cur-brm (command locateBall)) 0 current-wrlds))
   
   ;; The rest of the actions don't sense anything so they don't
   ;; prune the possible worlds list either.
   ((equal action 'domains::grab-ball)
    (if (command grabBall) ; If action succeded
        (progn
          ;; Update expected state
          (setf cur-rrm cur-brm) ; Update current robot room
          (setf cur-h Y) ; and holding status
          (format t ":::: Post grab-ball action: ~A~A~A~%" cur-rrm cur-brm cur-h)
          ;; And check to see if that's where we are
          (if (and (= cur-h (setf cur-h (command holdingBall)))
                   (= cur-rrm (setf cur-rrm (command inGreenRm))))
              current-wrlds ; everything is as we expect it
              (progn (setf cur-brm U) r) ; replanning if nessisary
            )
          )
      r ; If the action failed, also replan
      )
    )
   
   ((equal action 'domains::release-ball)
    (if (command releaseBall)
        (progn 
          (setf cur-h N)
          (format t ":::: Post release-ball action: ~A~A~A~%" cur-rrm cur-brm cur-h)
          (if (= cur-h (setf cur-h (command holdingBall)))
              current-wrlds
              r 
            )
          )
        r
      )
    )
   
   ((equal action 'domains::face-dest)
    (if (command locateDest)  ; Note locateDest on the robot is called
        current-wrlds         ; face dest in the planner, I should go
        r                     ; through and rename things.          
      )
    )
   
   
   ((equal action 'domains::go-dest)
    (if (command goRoom)
        (progn
          ;; Update expected state
          (cond
           ((= cur-rrm Y) (setf cur-rrm N))
           ((= cur-rrm N) (setf cur-rrm Y))
           )
          (if (= cur-h Y) (setf cur-brm cur-rrm))
          (format t ":::: Post go-dest action: ~A~A~A~%" cur-rrm cur-brm cur-h)
          ;; Check to see if things are as we expect.
          (if (and (= cur-h (setf cur-h (command holdingBall)))
                   (= cur-rrm (setf cur-rrm (command inGreenRm))))
              current-wrlds ; everything is as we expect it
              (progn (setf cur-brm U) r) ; things not as we expect
            )
          )
        r ; Action failed
      )
    )
   ) ; End cond
  ) ; end defun

;;; Some handy accessors for parts of plans
(defun getaction (pplan) (caar pplan))
(defun get-noex-wrlds (pplan) (cdr pplan))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Gen worlds function and auxillaries     ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; This table is used to ensure that the generated
;; possible worlds list is in the same order as the contexts
;; generated by SGP
(defvar worldstable (list
                     (vector y y y) ; Context 1 (in UUU)
                     (vector y y n) ; 2
                     (vector y n n) ; 3
                     (vector n y n) ; 4
                     (vector n n y) ; 5
                     (vector n n n) ; 6
                     )
  )

;;; Accessors for the data returned by getworlds
;;; and prune worlds
(defun getworld-no (world) (car world))
(defun getstate-vec (world) (cdr world))

(defun getworlds (rrm brm h)
  (gwaux rrm brm h worldstable 1 ()))

(defun gwaux (r b h table ind retval)
  (cond
   ((null table) retval)
   ((matches-world r b h (car table)) 
    (gwaux r b h (cdr table) (+ 1 ind) (cons (cons ind (car table)) retval)))
   (t (gwaux r b h (cdr table) ind retval))
   )
  )

(defun matches-world (sr sb sh world)
  (let ((wr (svref world 0))
        (wb (svref world 1))
        (wh (svref world 2))
        )
    (and (or (= sr u) (= sr wr))
         (or (= sb u) (= sb wb))
         (or (= sh u) (= sh wh))
         )
    )
  )

(defun prune-worlds (r-mask b-mask h-mask world-list)
  (prunewaux r-mask b-mask h-mask world-list ()))

(defun prunewaux (r b h lst retval)
  (cond
   ((null lst) retval) ; No more worlds to check through
   ((matches-world r b h (getstate-vec (car lst)))
    ;; If the world matches the mask it isn't pruned
    (prunewaux r b h (cdr lst) (cons (car lst) retval)))
   (t (prunewaux r b h (cdr lst) retval)) ; Else don't at the world and keep looking
   )
  )