//-*-c++-*-
#ifndef PLAN_MANAGER_h_
#define PLAN_MANAGER_h_

#include "Behaviors/BehaviorBase.h"
#include <iostream>
#include "Wireless/Wireless.h"
#include "Shared/ProjectInterface.h"

// Define tirnary predicate values
#define YES 1
#define NO -1
#define UNKNOWN 0

class PlanManager : public BehaviorBase {
public:
	//! constructor
	PlanManager();

	//! Destructor
	~PlanManager();

	virtual void DoStart();

	virtual void DoStop();

	virtual void processEvent(const EventBase& e);

	virtual void runCommand(const std::string& s);
	
	static int aiComm(char *buf, int bytes); //called by wireless when there's new data

	static std::string getClassDescription() { return "The manager behavior for planning"; }
	virtual std::string getDescription() const { return getClassDescription(); }
	
	enum ACTIONS { LOCATE_BALL = 0,
		           CHECK_ROOM = 1,
		           LOCATE_DEST = 2,
		           CHECK_HOLDING = 3,
		           COLOCATE = 4,
		           GRAB_BALL = 5,
		           GO_OTHER_RM = 6,
		           RELEASE_BALL = 7,
	};
	
	static const unsigned int EFFECT_CHECK = 100;
	
	virtual void runAction(ACTIONS act);
	
	static const unsigned int timeOutSID = 1001;
	
	std::string printActionName(unsigned int action);
	void printState();
	
	// Keeping track of important locations in offset coordinates
	//float ballX, ballY, destX, destY, robotA;
	float destDist, destBearing;
	// #defined YES, NO and UNKNOWN are encoded in these vars
	int inGreenRm, ballInGreenRm, holdingBall;

	bool actionFailed;

protected:
	BehaviorBase* actionArray[8];
	
	//! The input command stream socket
	Socket *cmdsock;
	
	bool doingCheck;
private:
	/* Processes locomotion events to keep track of the location of the destination
	 * and ball in egocentric coordinates */
	//void updateMap(const LocomotionEvent* lEvent);
	//const LocomotionEvent* lastMot;
	
	void lieDown(); // Lies the AIBO down and turns off the headlight

	// Providing declarations for these functions will avoid a compiler warning if
	// you have any class members which are pointers.  However, as it is, an error
	// will result if you inadvertantly cause a call to either (which is probably
	// a good thing, unless you really intended to copy/assign a behavior, in
	// which case simply provide implementations for the functions)
	PlanManager(const PlanManager&); //!< don't call (copy constructor)
	PlanManager& operator=(const PlanManager&); //!< don't call (assignment operator)
};

//Global pointer to the manager. Much like erouter
extern PlanManager * manager;
static const float headHoldBallTilt = -1.02;

#endif
